import asyncio
import logging
import re
import sys
from collections import defaultdict
from concurrent.futures import Executor
from types import MappingProxyType
from typing import (
    Any, DefaultDict, Dict, Mapping, Optional, Sequence, Set, Tuple,
)

from .base import Service


try:
    import grpc.aio
    from grpc_reflection.v1alpha import reflection
except ImportError as e:
    raise ImportError(
        "You must install 'grpcio' manually or using extras 'aiomisc[grpc]'",
    ) from e

log = logging.getLogger(__name__)

if sys.version_info >= (3, 9):
    PortFuture = asyncio.Future[int]
else:
    PortFuture = asyncio.Future


class GRPCService(Service):
    GRACEFUL_STOP_TIME: float = 60.

    _ADDRESS_REGEXP = re.compile(
        r"(?P<address>(\[((([([0-9a-fA-F:]*)+)])?|([\w.]+))):(\d+)",
    )

    _server: grpc.aio.Server
    _server_args: MappingProxyType
    _insecure_ports: Set[Tuple[str, PortFuture]]
    _secure_ports: Set[Tuple[str, grpc.ServerCredentials, PortFuture]]
    _registered_services: DefaultDict[str, Dict[str, grpc.RpcMethodHandler]]

    def __init__(
        self, *,
        migration_thread_pool: Optional[Executor] = None,
        handlers: Optional[Sequence[grpc.ServiceRpcHandler]] = None,
        interceptors: Optional[Sequence[Any]] = None,
        options: Optional[Sequence[Tuple[str, Any]]] = None,
        maximum_concurrent_rpcs: Optional[int] = None,
        compression: Optional[grpc.Compression] = None,
        reflection: bool = False,
        **kwds: Any,
    ):
        self._server_args = MappingProxyType({
            "compression": compression,
            "handlers": handlers,
            "interceptors": interceptors,
            "maximum_concurrent_rpcs": maximum_concurrent_rpcs,
            "migration_thread_pool": migration_thread_pool,
            "options": options,
        })
        self._services: Set[grpc.ServiceRpcHandler] = set()
        self._insecure_ports = set()
        self._secure_ports = set()
        self._reflection = reflection
        self._registered_services = defaultdict(dict)
        super().__init__(**kwds)

    @classmethod
    def _log_port(cls, msg: str, address: str, bind_port: Any) -> None:
        match: Optional[re.Match] = cls._ADDRESS_REGEXP.match(address)

        if match is not None:
            groups = match.groupdict()
            address = groups["address"]

        log.info("%s: grpc://%s:%s", msg, address, bind_port)

    async def start(self) -> None:
        self._server = grpc.aio.server(**self._server_args)

        for address, future in self._insecure_ports:
            port = self._server.add_insecure_port(address)
            future.set_result(port)
            self._log_port("Listening insecure address", address, port)

        for address, credentials, future in self._secure_ports:
            port = self._server.add_secure_port(address, credentials)
            future.set_result(port)
            self._log_port("Listening secure address", address, port)

        if self._reflection:
            service_names = [x.service_name() for x in self._services]
            service_names.append(reflection.SERVICE_NAME)
            reflection.enable_server_reflection(service_names, self._server)

        for name, handlers in self._registered_services.items():
            # noinspection PyUnresolvedReferences
            self._server.add_registered_method_handlers(   # type: ignore
                name, handlers,
            )

        self._server.add_generic_rpc_handlers(tuple(self._services))
        await self._server.start()

    async def stop(self, exception: Optional[Exception] = None) -> None:
        await self._server.stop(self.GRACEFUL_STOP_TIME)

    def add_generic_rpc_handlers(
        self, generic_rpc_handlers: Sequence[grpc.ServiceRpcHandler],
    ) -> None:
        for service in generic_rpc_handlers:
            self._services.add(service)

    def add_registered_method_handlers(
        self, name: str, handlers: Mapping[str, grpc.RpcMethodHandler],
    ) -> None:
        self._registered_services[name].update(handlers)

    def add_insecure_port(self, address: str) -> PortFuture:
        future: PortFuture = asyncio.Future()
        self._insecure_ports.add((address, future))
        return future

    def add_secure_port(
        self, address: str,
        server_credentials: grpc.ServerCredentials,
    ) -> PortFuture:
        future: PortFuture = asyncio.Future()
        self._secure_ports.add((address, server_credentials, future))
        return future



from typing import Any, Dict, Generic, Hashable, List, Optional, Tuple, TypeVar


K = Tuple[str, ...]
T = TypeVar("T", bound=Any)


class RadixNode(Generic[T]):
    __slots__ = ("children", "value")

    def __init__(self) -> None:
        self.children: Dict[Hashable, RadixNode[T]] = {}
        self.value: Optional[T] = None


class RadixTree(Generic[T]):
    root: RadixNode[T]

    __slots__ = ("root",)

    def __init__(self) -> None:
        self.root = RadixNode()

    def insert(self, key: K, value: Optional[T]) -> None:
        node = self.root
        for part in key:
            if part not in node.children:
                node.children[part] = RadixNode()
            node = node.children[part]
        node.value = value

    def search(self, key: K) -> Optional[T]:
        node = self.root
        for part in key:
            if part not in node.children:
                return None
            node = node.children[part]
        return node.value

    def find_prefix(self, key: K) -> Optional[Tuple[K, T]]:
        node = self.root
        longest_prefix: List[str] = []
        value = None
        part: Hashable
        if node.value is not None:
            value = node.value
        for part in key:
            if part in node.children:
                node = node.children[part]
                longest_prefix.append(part)
                if node.value is not None:
                    value = node.value
            else:
                break

        if value is None:
            return None

        return tuple(longest_prefix), value



import asyncio
import atexit
import logging
import logging.handlers
import time
import traceback
from contextlib import suppress
from functools import partial
from socket import socket
from typing import (
    Any, Callable, Dict, Iterable, List, Optional, Tuple, Type, Union,
)
from weakref import finalize

import aiomisc_log
from aiomisc_log.enum import LogFormat, LogLevel

from .thread_pool import run_in_new_thread


def _thread_flusher(
    handler: logging.handlers.MemoryHandler,
    flush_interval: Union[float, int],
    loop: asyncio.AbstractEventLoop,
) -> None:
    def has_no_target() -> bool:
        return True

    def has_target() -> bool:
        return bool(handler.target)

    is_target = has_no_target

    if isinstance(handler, logging.handlers.MemoryHandler):
        is_target = has_target

    while not loop.is_closed() and is_target():
        with suppress(Exception):
            if handler.buffer:
                handler.flush()

        time.sleep(flush_interval)


def suppressor(
    callback: Callable[..., None],
    exceptions: Tuple[Type[BaseException], ...] = (Exception,),
) -> Callable[..., None]:
    def wrapper() -> None:
        with suppress(*exceptions):
            callback()
    return wrapper


def wrap_logging_handler(
    handler: logging.Handler,
    loop: Optional[asyncio.AbstractEventLoop] = None,
    buffer_size: int = 1024,
    flush_interval: Union[float, int] = 0.1,
) -> logging.Handler:
    buffered_handler = logging.handlers.MemoryHandler(
        buffer_size,
        target=handler,
        flushLevel=logging.CRITICAL,
    )

    run_in_new_thread(
        _thread_flusher, args=(
            buffered_handler, flush_interval, loop,
        ), no_return=True, statistic_name="logger",
    )

    at_exit_flusher = suppressor(handler.flush)
    atexit.register(at_exit_flusher)
    finalize(buffered_handler, partial(atexit.unregister, at_exit_flusher))

    return buffered_handler


class UnhandledLoopHook(aiomisc_log.UnhandledHook):
    @staticmethod
    def _fill_transport_extra(
        transport: Optional[asyncio.Transport],
        extra: Dict[str, Any],
    ) -> None:
        if transport is None:
            return

        extra["transport"] = repr(transport)

        for key in (
            "peername", "sockname", "compression",
            "cipher", "peercert", "pipe", "subprocess",
        ):
            value = transport.get_extra_info(key)
            if value:
                extra[f"transport_{key}"] = value

    def __call__(
        self, loop: asyncio.AbstractEventLoop,
        context: Dict[str, Any],
    ) -> None:
        context = dict(context)
        message: str = context.pop("message", "unhandled loop exception")
        exception: Optional[BaseException] = context.pop("exception", None)
        future: Optional[asyncio.Future] = context.pop("future", None)
        task: Optional[asyncio.Task] = context.pop("task", None)
        handle: Optional[asyncio.Handle] = context.pop("handle", None)
        protocol: Optional[asyncio.Protocol] = context.pop("protocol", None)
        transport: Optional[asyncio.Transport] = context.pop("transport", None)
        sock: Optional[socket] = context.pop("socket", None)
        source_traceback: List[traceback.FrameSummary] = (
            context.pop("source_traceback", None) or []
        )

        if exception is None:
            if future is not None:
                exception = future.exception()
            elif task is not None and task.done():
                exception = task.exception()

        extra = context
        if handle is not None:
            extra["handle"] = repr(handle)
        if protocol is not None:
            extra["protocol"] = repr(protocol)
        if sock is not None:
            extra["sock"] = repr(sock)

        self._fill_transport_extra(transport, extra)
        self.logger.exception(message, exc_info=exception, extra=extra)
        if source_traceback:
            self.logger.error(
                "".join(traceback.format_list(source_traceback)),
            )


def basic_config(
    level: Union[int, str] = LogLevel.default(),
    log_format: Union[str, LogFormat] = LogFormat.default(),
    buffered: bool = True, buffer_size: int = 1024,
    flush_interval: Union[int, float] = 0.2,
    loop: Optional[asyncio.AbstractEventLoop] = None,
    handlers: Iterable[logging.Handler] = (),
    **kwargs: Any,
) -> None:
    loop = loop or asyncio.get_event_loop()
    unhandled_hook = UnhandledLoopHook(logger_name="asyncio.unhandled")

    def wrap_handler(handler: logging.Handler) -> logging.Handler:
        nonlocal buffer_size, buffered, loop, unhandled_hook

        unhandled_hook.add_handler(handler)

        if buffered:
            return wrap_logging_handler(
                handler=handler,
                buffer_size=buffer_size,
                flush_interval=flush_interval,
                loop=loop,
            )
        return handler

    aiomisc_log.basic_config(
        level=level,
        log_format=log_format,
        handler_wrapper=wrap_handler,
        handlers=handlers,
        **kwargs,
    )

    loop.set_exception_handler(unhandled_hook)


__all__ = (
    "LogFormat",
    "LogLevel",
    "basic_config",
)



from .base import Service, ServiceMeta, SimpleServer
from .process import ProcessService, RespawningProcessService
from .profiler import Profiler
from .tcp import RobustTCPClient, TCPClient, TCPServer
from .tls import RobustTLSClient, TLSClient, TLSServer
from .tracer import MemoryTracer
from .udp import UDPServer


__all__ = (
    "MemoryTracer",
    "ProcessService",
    "Profiler",
    "RespawningProcessService",
    "RobustTCPClient",
    "RobustTLSClient",
    "Service",
    "ServiceMeta",
    "SimpleServer",
    "TCPClient",
    "TCPServer",
    "TLSClient",
    "TLSServer",
    "UDPServer",
)



import logging
from typing import IO, Any, Optional

from aiomisc_log.enum import DateFormat


try:
    from rich import reconfigure
    from rich.logging import RichHandler

    def rich_formatter(
        date_format: Optional[str] = None, stream: Optional[IO[str]] = None,
        **kwargs: Any,
    ) -> logging.Handler:
        kwargs.setdefault("rich_tracebacks", False)
        kwargs.setdefault(
            "log_time_format", date_format or DateFormat.rich.value,
        )

        if "console" not in kwargs:
            if stream is None:
                reconfigure(stderr=True)
            else:
                reconfigure(file=stream)

        handler = RichHandler(**kwargs)
        formatter = logging.Formatter("%(message)s")
        handler.setFormatter(formatter)
        return handler
except ImportError:
    def rich_formatter(
        date_format: Optional[str] = None, stream: Optional[IO[str]] = None,
        **kwargs: Any,
    ) -> logging.Handler:
        raise ImportError("You must install \"rich\" library for use it")



import asyncio
import sys
from functools import wraps
from typing import Any, Callable, Coroutine, TypeVar, Union


if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec


T = TypeVar("T")
P = ParamSpec("P")
Number = Union[int, float]


def timeout(
    value: Number
) -> Callable[
    [Callable[P, Coroutine[Any, Any, T]]],
    Callable[P, Coroutine[Any, Any, T]],
]:
    def decorator(
        func: Callable[P, Coroutine[Any, Any, T]],
    ) -> Callable[P, Coroutine[Any, Any, T]]:
        if not asyncio.iscoroutinefunction(func):
            raise TypeError("Function is not a coroutine function")

        @wraps(func)
        async def wrap(*args: P.args, **kwargs: P.kwargs) -> T:
            return await asyncio.wait_for(
                func(*args, **kwargs),
                timeout=value,
            )
        return wrap
    return decorator



from . import io, log
from ._context_vars import StrictContextVar
from .aggregate import aggregate, aggregate_async
from .backoff import asyncbackoff, asyncretry
from .circuit_breaker import CircuitBreaker, CircuitBroken, cutout
from .context import Context, get_context
from .counters import Statistic, get_statistics
from .entrypoint import CURRENT_ENTRYPOINT, Entrypoint, entrypoint, run
from .iterator_wrapper import IteratorWrapper
from .periodic import PeriodicCallback
from .plugins import plugins
from .pool import PoolBase
from .process_pool import ProcessPoolExecutor
from .recurring import (
    RecurringCallback, StrategyException, StrategySkip, StrategyStop,
)
from .service import Service
from .signal import Signal, receiver
from .thread_pool import (
    IteratorWrapperSeparate, ThreadPoolExecutor, context_partial, sync_await,
    sync_wait_coroutine, threaded, threaded_iterable,
    threaded_iterable_separate, threaded_separate, wait_coroutine,
)
from .timeout import timeout
from .utils import (
    SelectResult, awaitable, bind_socket, cancel_tasks, chunk_list,
    new_event_loop, select, shield,
)
from .version import __version__, version_info
from .worker_pool import WorkerPool


__all__ = (
    "CURRENT_ENTRYPOINT",
    "CircuitBreaker",
    "CircuitBroken",
    "Context",
    "Entrypoint",
    "IteratorWrapper",
    "IteratorWrapperSeparate",
    "PeriodicCallback",
    "PoolBase",
    "ProcessPoolExecutor",
    "RecurringCallback",
    "SelectResult",
    "Service",
    "Signal",
    "Statistic",
    "StrategyException",
    "StrategySkip",
    "StrategyStop",
    "StrictContextVar",
    "ThreadPoolExecutor",
    "WorkerPool",
    "__version__",
    "aggregate",
    "aggregate_async",
    "asyncbackoff",
    "asyncretry",
    "awaitable",
    "bind_socket",
    "cancel_tasks",
    "chunk_list",
    "context_partial",
    "cutout",
    "entrypoint",
    "get_context",
    "get_statistics",
    "io",
    "log",
    "new_event_loop",
    "plugins",
    "receiver",
    "run",
    "select",
    "shield",
    "sync_await",
    "sync_wait_coroutine",
    "threaded",
    "threaded_iterable",
    "threaded_iterable_separate",
    "threaded_separate",
    "timeout",
    "version_info",
    "wait_coroutine",
)



from . import records
from .service import DNSServer, TCPDNSServer, UDPDNSServer
from .store import DNSStore
from .zone import DNSZone


__all__ = (
    "DNSServer",
    "DNSStore",
    "DNSZone",
    "TCPDNSServer",
    "UDPDNSServer",
    "records",
)



from typing import Optional, Sequence, Tuple

from .records import DNSRecord, RecordType
from .tree import RadixTree
from .zone import DNSZone


class DNSStore:
    zones: RadixTree[DNSZone]

    __slots__ = ("zones",)

    def __init__(self) -> None:
        self.zones = RadixTree()

    def add_zone(self, zone: DNSZone) -> None:
        zone_tuple = self.get_reverse_tuple(zone.name)
        if self.zones.search(zone_tuple):
            raise ValueError(f"Zone {zone.name} already exists.")
        self.zones.insert(zone_tuple, zone)

    def remove_zone(self, zone_name: str) -> None:
        zone_tuple = self.get_reverse_tuple(zone_name)
        if not self.zones.search(zone_tuple):
            raise ValueError(
                f"Zone {zone_name} does not exist.",
            )
        # Clear zone from RadixTree
        self.zones.insert(zone_tuple, None)

    def get_zone(self, zone_name: str) -> Optional[DNSZone]:
        zone_tuple = self.get_reverse_tuple(zone_name)
        return self.zones.search(zone_tuple)

    def query(self, name: str, record_type: RecordType) -> Sequence[DNSRecord]:
        if not name.endswith("."):
            name += "."
        zone_tuple = self.get_zone_for_name(name)
        if not zone_tuple:
            return ()
        zone = self.zones.search(zone_tuple)
        return zone.get_records(name, record_type) if zone is not None else ()

    def get_zone_for_name(self, name: str) -> Optional[Tuple[str, ...]]:
        labels = self.get_reverse_tuple(name)
        result = self.zones.find_prefix(labels)
        return result[0] if result else None

    @staticmethod
    def get_reverse_tuple(zone_name: str) -> Tuple[str, ...]:
        return tuple(zone_name.strip(".").split("."))[::-1]



from collections import defaultdict
from typing import DefaultDict, Sequence, Set, Tuple

from .records import DNSRecord, RecordType


class DNSZone:
    records: DefaultDict[Tuple[str, RecordType], Set[DNSRecord]]
    name: str

    __slots__ = ("name", "records")

    def __init__(self, name: str):
        if not name.endswith("."):
            name += "."
        self.name = name
        self.records = defaultdict(set)

    def add_record(self, record: DNSRecord) -> None:
        if not self._is_valid_record(record):
            raise ValueError(
                f"Record {record.name} does not belong to zone {self.name}",
            )
        key = (record.name, record.type)
        self.records[key].add(record)

    def remove_record(self, record: DNSRecord) -> None:
        key = (record.name, record.type)
        if key in self.records:
            self.records[key].discard(record)
            if self.records[key]:
                return
            del self.records[key]

    def get_records(
        self, name: str, record_type: RecordType,
    ) -> Sequence[DNSRecord]:
        if not name.endswith("."):
            name += "."
        key = (name, record_type)
        return tuple(self.records.get(key, ()))

    def _is_valid_record(self, record: DNSRecord) -> bool:
        return record.name.endswith(self.name)
